#include <ostream>
#include <sstream>
#include <cmath>

#include "expr.hh"
#include "stmt.hh"
#include "typenameof.hh"
#include "compile.hh"
#include "dprintf.hh"
#include "util.hh"

expr::expr()
{
}
expr::~expr()
{
}

//static const std::string TypeToString(TypeP p)
//{
//    return p->Dump();
//}

tristate expr::is_num() const
{
    TypeP t = GetType();
#ifdef BINDING_DEBUG
    //dprintf(dp_declare, "is_num? %s\n", t?TypeToString(t).c_str():"NULL");
#endif
    if(!t) return maybe; // unknown
    return !t->is_array() && t->is_num();
}

tristate expr::is_string() const
{
    TypeP t = GetType();
#ifdef BINDING_DEBUG
    //dprintf(dp_declare, "is_string? %s\n", t?TypeToString(t).c_str():"NULL");
#endif
    if(!t) return maybe; // unknown
    return !t->is_array() && t->is_string();
}

bool expr_number::is_int() const
{
    if(!type) return false;
    if(type->is_array()) return false;
    return type->Type == TypeType::Integer
        || type->Type == TypeType::Long
        || type->Type == TypeType::Byte
        || type->Type == TypeType::UnsignedInteger
        || type->Type == TypeType::UnsignedLong
        || type->Type == TypeType::UnsignedByte
        || float_val == int_val;
}
long double   expr_number::get_f() const { return is_int() ? int_val : float_val; }
long long int expr_number::get_i() const { return is_int() ? int_val : (int)std::floor(float_val + 0.5); }

TypeP expr_unary::GetType() const
{
    TypeP t = p->GetType();
    if(t && token == NOT)
        switch(t->Type)
        {
            case TypeType::Byte:
            case TypeType::UnsignedByte:
            case TypeType::UnsignedInteger:
            case TypeType::UnsignedLong:
            case TypeType::Long:         return t;
            default: return TypeInt;
        }
    return t;
}

expr* expr_unary::optimized() &&
{
    expr_number* n = dynamic_cast<expr_number*> (&*p);
    switch(token)
    {
        case MINUS: if(n) return (new expr_number(- n->get_i())); break;
        case NOT:   if(n) return (new expr_number(~ n->get_i())); break;
        default: break;
    }
    return std::move(*this).expr::optimized();
}

static int IntegerLevel(TypeType type)
{
    // Unsigned types are EVEN
    switch(type)
    {
        case TypeType::UnsignedLong:    return 6;
        case TypeType::Long:            return 5;
        case TypeType::UnsignedInteger: return 4;
        case TypeType::Integer:         return 3;
        case TypeType::UnsignedByte:    return 2;
        case TypeType::Byte:            return 1;

        case TypeType::UnsignedBit:
        case TypeType::Bit:
            // TODO
            break;

        case TypeType::UserType:
        case TypeType::Single:
        case TypeType::Double:
        case TypeType::String:
            break;
    }
    return -1;
}
static TypeP TypeForLevel(int level)
{
    switch(level)
    {
        case 6: return TypeUlong;
        case 5: return TypeLng;
        case 4: return TypeUint;
        case 3: return TypeInt;
        case 2: return TypeUint;
        case 1: return TypeByte;
    }
    return nullptr;
}

TypeP expr_binary::GetType() const
{
    // choices:
    //   OR, XOR, AND, comparisons, plus, minus, MOD, '\\', mul, div, '^'

    // Is the output an integer?
    bool always_signed_result   = false;
    bool never_promote          = false;
    bool always_unsigned_result = false;
    bool always_integer         = false;
    bool maybe_integer          = false;

    switch(token)
    {
        case OR: case XOR: case AND:
        case qbcSHR:
            never_promote  = true;
            always_integer = true;
            break;

        case GREATERTHAN: case LESSTHAN: case EQUALS:
        case NOTEQUAL: case LESSOREQUAL: case GREATEROREQUAL:
            always_signed_result = true;
            always_integer       = true;
            break;

        case MOD:
            always_unsigned_result = true;
            always_integer         = true;
            break;

        case BKSLASH:
        case qbcSHL:
            always_integer = true;
            break;

        case PLUS: case MINUS: case ASTERISK:
            maybe_integer  = true;
            break;

        default: break;
    }
    TypeP a = p0->GetType(); if(!a) return a;
    TypeP b = p1->GetType(); if(!b) return b;

    int intlevel_a = IntegerLevel(a->Type);
    int intlevel_b = IntegerLevel(b->Type);
    int intlevel_min = std::min(intlevel_a, intlevel_b);
    int intlevel_max = std::max(intlevel_a, intlevel_b);
    //std::cout << "/* atype=" << a->Dump() << ", btype=" << b->Dump() << " */";

    if(always_integer
    || (maybe_integer && intlevel_min >= 1))
    {
        int return_type = intlevel_max;
        if(!never_promote && return_type < 3) return_type = 3;
        if(return_type < 1) return_type = 1;
        return TypeForLevel(return_type);
    }

    if(a->Type == TypeType::String || b->Type == TypeType::String) return TypeStr;
    if(a->Type == TypeType::Single && b->Type == TypeType::Single) return TypeSng;
    if(a->Type == TypeType::UserType || b->Type == TypeType::UserType) return 0;
    return TypeDbl;
}

bool expr_binary::IsComparison() const
{
    switch(token)
    {
        case LESSTHAN:
        case GREATERTHAN:
        case EQUALS:
        case LESSOREQUAL:
        case GREATEROREQUAL:
        case NOTEQUAL:
            return true;
        case AND:
        case OR:
            return p0->IsComparison() || p1->IsComparison();
        default: break;
    }
    return false;
}

expr* expr_binary::optimized() &&
{
    expr_string* s0 = dynamic_cast<expr_string*> (&*p0);
    expr_string* s1 = dynamic_cast<expr_string*> (&*p1);
    switch(token)
    {
        #define stringop(t,op,p) if(s0&&s1) return (new t((s0->get() op s1->get()) p)); break
        case PLUS:           stringop(expr_string,+,);
        case LESSTHAN:       stringop(expr_number,<,?-1:0);
        case GREATERTHAN:    stringop(expr_number,>,?-1:0);
        case EQUALS:         stringop(expr_number,==,?-1:0);
        case LESSOREQUAL:    stringop(expr_number,<=,?-1:0);
        case GREATEROREQUAL: stringop(expr_number,>=,?-1:0);
        case NOTEQUAL:       stringop(expr_number,!=,?-1:0);
        #undef stringop
        default: break;
    }

    bool commutable = (token == PLUS || token == ASTERISK);
    if(commutable)
    {
        /* If p1 is SQR and p0 is not, swap them, so SQR comes first.
         * This helps the ASINH/ACOSH optimization.
         * Do the same for POW and EXP.
         */
        auto p0f = dynamic_cast<expr_builtin_numeric*> (&*p0);
        if(auto p1f = dynamic_cast<expr_builtin_numeric*> (&*p1))
            if( (p1f->GetToken() == SQR || p1f->GetToken() == EXP || p1f->GetToken() == CIRCUMFLEX)
            && (!p0f
            || !(p0f->GetToken() == SQR || p0f->GetToken() == EXP || p0f->GetToken() == CIRCUMFLEX)))
            {
                std::swap(p0, p1);
                return std::move(*this).optimized();
            }

        /* If p0 and p1 are identical, change this into
         * a multiplication / exponentation by 2.
         * This helps gcc in some cases, as it has very efficient
         * code generation for integer-powis.
         */
        if(p0->IsSameAs(*p1))
        {
            token = (token == PLUS ? ASTERISK : CIRCUMFLEX);
            p1.reset( new expr_number(2, TypeUbyte) );
            return std::move(*this).optimized();
        }
    }

    expr_number* n0 = dynamic_cast<expr_number*> (&*p0);
    expr_number* n1 = dynamic_cast<expr_number*> (&*p1);
    switch(token)
    {
        #define numberop(t,op,p) if(n0&&n1) return (new t((n0->get_f() op n1->get_f()) p))
        case PLUS:
        {
            numberop(expr_number,+,);

            // Change exp(x) + exp(-x) to cosh(x)*2 if possible
            auto ep0 = dynamic_cast<expr_builtin_function*>(&*get_p0());
            auto ep1 = dynamic_cast<expr_builtin_function*>(&*get_p1());
            if(ep0 && ep1 && ep0->GetToken() == EXP && ep1->GetToken() == EXP)
            {
                auto u1 = dynamic_cast<expr_unary*> (&*ep1->GetParam(0));
                auto u0 = dynamic_cast<expr_unary*> (&*ep0->GetParam(0));
                if(u1 && u1->GetToken() == MINUS && ep0->GetParam(0)->IsSameAs(*u1->p))
                {
                    token = ASTERISK;
                    ep0->ChangeToken( qbcCOSH );
                    p1.reset( new expr_number(2, TypeUbyte) );
                    return std::move(*this).optimized();
                }
                if(u0 && u0->GetToken() == MINUS && ep1->GetParam(0)->IsSameAs(*u0->p))
                {
                    token = ASTERISK;
                    ep1->ChangeToken( qbcCOSH );
                    p0.reset( new expr_number(2, TypeUbyte) );
                    return std::move(*this).optimized();
                }
            }
        } break;

        case MINUS:
        {
            numberop(expr_number,-,);

            // Change exp(x) - exp(-x) to sinh(x)*2 if possible
            auto ep0 = dynamic_cast<expr_builtin_function*>(&*get_p0());
            auto ep1 = dynamic_cast<expr_builtin_function*>(&*get_p1());
            if(ep0 && ep1 && ep0->GetToken() == EXP && ep1->GetToken() == EXP)
            {
                auto u = dynamic_cast<expr_unary*> (&*ep1->GetParam(0));
                if(u && u->GetToken() == MINUS && ep0->GetParam(0)->IsSameAs(*u->p))
                {
                    token = ASTERISK;
                    ep0->ChangeToken( qbcSINH );
                    p1.reset( new expr_number(2, TypeUbyte) );
                    return std::move(*this).optimized();
                }
            }
        } break;

        case SLASH:
        {
            // Avoid division by zero
            if(n0&&n1 && n1->get_f() != 0.) return new expr_number((n0->get_f() / n1->get_f()));

            // Check if we have (exp(x)-1) / (exp(x)+1), which can be converted into tanh(x*0.5)
            auto bp0 = dynamic_cast<expr_binary*>(&*get_p0());
            auto bp1 = dynamic_cast<expr_binary*>(&*get_p1());
            if(bp0 && bp1 && (bp0->GetToken() == MINUS || bp0->GetToken() == PLUS)
                          && bp1->GetToken() == PLUS)
            {
                // We've got (x + y) / (c + d)  or   (x - y) / (c + d)
                auto bp0p0 = dynamic_cast<expr_builtin_function*>(&*bp0->get_p0());
                auto bp1p0 = dynamic_cast<expr_builtin_function*>(&*bp1->get_p0());
                auto bp0p1 = dynamic_cast<expr_number*>(&*bp0->get_p1());
                auto bp1p1 = dynamic_cast<expr_number*>(&*bp1->get_p1());
                if(bp0p0 && bp1p0 && bp0p0->GetToken() == EXP && bp0p0->IsSameAs(*bp1p0)
                && bp0p1 && bp1p1 && bp1p1->get_f() == 1)
                {
                    // We've got (EXP(x) + y) / (EXP(x) + 1)  or  (EXP(x) - y) / (EXP(x) + 1)
                    if((bp0->GetToken() == MINUS && bp0p1->get_f() == 1)
                    || (bp0->GetToken() == PLUS  && bp0p1->get_f() == -1))
                    {
                        exprp subexpr_p0 = std::move(bp0p0->GetParam(0));
                        exprp subexpr ( new expr_binary(ASTERISK, std::move(subexpr_p0), exprp(new expr_number(0.5, bp0p0->GetType()))) );
                        OptimizeExprp(subexpr);
                        exprp result ( new expr_builtin_numeric(LexLocation(), qbcTANH, std::move(subexpr), "N") );
                        OptimizeExprp(result);
                        return result.release();
            }   }   }
        } break;

        case ASTERISK:
        {
            // Change a * (1 SHL b)  into a SHL b
            expr_binary* b1 = dynamic_cast<expr_binary*> (&*p1);
            if(b1 && b1->token == qbcSHL)
            {
                expr_number *bn0 = dynamic_cast<expr_number*> (&*b1->p0);
                if(bn0 && bn0->is_int() && bn0->get_i() == 1)
                {
                    // Move their p1 here
                    exprp tmp = std::move(b1->p1); p1 = std::move(tmp);
                    token = qbcSHL;
                    // reoptimize
                    return std::move(*this).optimized();
            }   }

            // Change pow(x, n) * x into pow(x, n+1)
            expr_binary* b0 = dynamic_cast<expr_binary*> (&*p0);
            if(b0 && b0->GetToken() == CIRCUMFLEX && p1->IsSameAs(*b0->p0))
            {
                token = CIRCUMFLEX;
                std::swap(p0, p1); // p0 is now x, p1 is pow()
                p1.reset( new expr_binary(PLUS, std::move(b0->p1), exprp(new expr_number(1, TypeUbyte))) );
                OptimizeExprp(p1);
                return std::move(*this).optimized();
            }
            numberop(expr_number,*,);
        } break;

        case BKSLASH:
        {
            // Change  a / (1 SHL b)  into a SHR b

            expr_binary* b1 = dynamic_cast<expr_binary*> (&*p1);
            if(b1 && b1->token == qbcSHL)
            {
                expr_number *bn0 = dynamic_cast<expr_number*> (&*b1->p0);
                if(bn0 && bn0->is_int() && bn0->get_i() == 1)
                {
                    // Move their p1 here
                    exprp tmp = std::move(b1->p1); p1 = std::move(tmp);
                    token = qbcSHR;
                    // reoptimize
                    return std::move(*this).optimized();
                }
            }

            // Avoid division by zero
            if(n0&&n1 && n1->get_i() != 0) return (new expr_number( (int) (n0->get_i() / n1->get_i()))); // FIXME: How about longs?
        } break;

        case CIRCUMFLEX:
        {
            if(n0 && n0->is_int() && n0->get_i() == 2 && p1->GetType()->is_integer())
            {
                // Change 2^x into 1 << x
                p0.reset(new expr_number(1));
                token = qbcSHL;
                return std::move(*this).expr::optimized();
            }
            if(n0&&n1) return (new expr_number(std::pow(n0->get_f(), n1->get_f()))); break;
        } break;

        case LESSTHAN:       numberop(expr_number,<,?-1:0); break;
        case GREATERTHAN:    numberop(expr_number,>,?-1:0); break;
        case EQUALS:         numberop(expr_number,==,?-1:0); break;
        case LESSOREQUAL:    numberop(expr_number,<=,?-1:0); break;
        case GREATEROREQUAL: numberop(expr_number,>=,?-1:0); break;
        case NOTEQUAL:       numberop(expr_number,!=,?-1:0); break;

        case MOD: if(n0&&n1)
        {
            // FIXME: Avoid division by zero
            return (new expr_number(n0->get_i() % n1->get_i()));
        } break;

        case XOR: if(n0&&n1) return (new expr_number(n0->get_i() ^ n1->get_i())); break;
        case OR:  if(n0&&n1) return (new expr_number(n0->get_i() | n1->get_i())); break;
        case AND: if(n0&&n1) return (new expr_number(n0->get_i() & n1->get_i())); break;
        case qbcSHL: if(n0&&n1) return (new expr_number(n0->get_i() << n1->get_i())); break;
        case qbcSHR: if(n0&&n1) return (new expr_number(n0->get_i() >> n1->get_i())); break;

        #undef numberop

        default: break;
    }
    return std::move(*this).expr::optimized();
}

expr* expr_builtin_numeric::optimized() &&
{
    expr_number* n0 =
        params.size() >= 1
        ? dynamic_cast<expr_number*> (&*params[0])
        : nullptr;

    switch(token)
    {
        case ABS: if(n0) return (new expr_number(std::fabs(n0->get_f()))); break;
        case EXP: if(n0) return (new expr_number(std::exp(n0->get_f()))); break;
        case LOG: if(n0) return (new expr_number(std::log(n0->get_f())));
        {
            // If argument is "sqrt(x^2 + 1)) + x", we can convert this into asinh(x)
            // If argument is "sqrt(x^2 - 1)) + x", we can convert this into acosh(x)
            if(auto b = dynamic_cast<expr_binary*>(&*params[0]))
            {
                if(b->token == PLUS)
                {
                    // We've got x+y
                    if(auto p0 = dynamic_cast<expr_builtin_numeric*>(&*b->get_p0()))
                    if(p0->GetToken() == SQR)
                    {
                        // We've got x + sqrt(y)
                        if(auto p0b = dynamic_cast<expr_binary*>(&*p0->GetParam(0)))
                            if(p0b->GetToken() == PLUS || p0b->GetToken() == MINUS)
                            {
                                // We've got x + sqrt(y + z)  or  x + sqrt(y - z)
                                auto p0bp0 = dynamic_cast<expr_binary*>(&*p0b->get_p0());
                                auto p0bp1 = dynamic_cast<expr_number*>(&*p0b->get_p1());
                                if(p0bp0 && p0bp1 && p0bp0->GetToken() == CIRCUMFLEX
                                                  && p0bp0->get_p0()->IsSameAs(*b->get_p1()))
                                {
                                    // We've got x + sqrt(x^y + z)  or   x + sqrt(x^y - z)
                                    if(auto exp = dynamic_cast<expr_number*>(&*p0bp0->get_p1()))
                                        if(exp->get_f() == 2)
                                        {
                                            // We've got x + sqrt(x^2 + z)  or  x + sqrt(x^2 - z)
                                            if((p0b->GetToken() == PLUS && p0bp1->get_f() == 1)
                                            || (p0b->GetToken() ==MINUS && p0bp1->get_f() == -1))
                                            {
                                                token = qbcASINH;
                                                params[0] = std::move(b->get_p1());
                                                return std::move(*this).optimized();
                                            }
                                            else
                                            if((p0b->GetToken() == PLUS && p0bp1->get_f() == -1)
                                            || (p0b->GetToken() ==MINUS && p0bp1->get_f() == 1))
                                            {
                                                token = qbcACOSH;
                                                params[0] = std::move(b->get_p1());
                                                return std::move(*this).optimized();
                }   }       }   }       }   }
            // If argument is (1+x) / (1-x), we can convert this into atanh(x) * 2
                if(b->token == SLASH)
                {
                    auto bp0 = dynamic_cast<expr_binary*>(&*b->get_p0());
                    auto bp1 = dynamic_cast<expr_binary*>(&*b->get_p1());
                    if(bp0 && bp1 && bp0->GetToken() == PLUS && bp1->GetToken() == MINUS)
                    {
                        // We've got (a + b) / (c - d)
                        auto bp1p0 = dynamic_cast<expr_number*>(&*bp1->get_p0());
                        if(bp1p0 && bp1p0->get_f() == 1)
                        {
                            // We've got (a + b) / (1 - x)
                            if( (bp0->get_p0()->IsSameAs(*bp1->get_p1()) && bp0->get_p1()->IsSameAs(*bp1->get_p0()))
                            ||  (bp0->get_p1()->IsSameAs(*bp1->get_p1()) && bp0->get_p0()->IsSameAs(*bp1->get_p0())))
                            {
                                exprp atanh(new expr_builtin_numeric(std::move(lexloc), qbcATANH, std::move(bp1->get_p1()), "N"));
                                OptimizeExprp(atanh);

                                exprp result( new expr_binary(ASTERISK, std::move(atanh), exprp(new expr_number(2, TypeUbyte)) ) );
                                OptimizeExprp(result);

                                return result.release();
                }   }   }   }
            }
        } break;
        case SQR: if(n0) return (new expr_number(std::sqrt(n0->get_f()))); break;

        case qbcASIN: if(n0) return (new expr_number(std::asin(n0->get_f()))); break;
        case qbcACOS: if(n0) return (new expr_number(std::acos(n0->get_f()))); break;
        case SIN: if(n0) return (new expr_number(std::sin(n0->get_f()))); break;
        case COS: if(n0) return (new expr_number(std::cos(n0->get_f()))); break;
        case TAN: if(n0) return (new expr_number(std::tan(n0->get_f()))); break;
        case ATN: if(n0) return (new expr_number(std::atan(n0->get_f())));
        {
            // If argument is "x / sqrt(1 - x^2))", we can convert this into asin(x)
            if(auto b = dynamic_cast<expr_binary*>(&*params[0]))
                if(b->token == SLASH)
                {
                    // We've got x / y
                    if(auto p1 = dynamic_cast<expr_builtin_numeric*>(&*b->get_p1()))
                    if(p1->GetToken() == SQR)
                    {
                        // We've got x / sqrt(y)
                        if(auto p1b = dynamic_cast<expr_binary*>(&*p1->GetParam(0)))
                            if(p1b->GetToken() == MINUS)
                            {
                                // We've got x / sqrt(y - z)
                                auto p1bp0 = dynamic_cast<expr_number*>(&*p1b->get_p0());
                                auto p1bp1 = dynamic_cast<expr_binary*>(&*p1b->get_p1());
                                if(p1bp0 && p1bp1 && p1bp1->GetToken() == CIRCUMFLEX
                                                  && p1bp1->get_p0()->IsSameAs(*b->get_p0())
                                                  && p1bp0->get_f() == 1)
                                {
                                    // We've got x / sqrt(1 - x^z)
                                    if(auto exp = dynamic_cast<expr_number*>(&*p1bp1->get_p1()))
                                        if(exp->get_f() == 2)
                                        {
                                            token = qbcASIN;
                                            params[0] = std::move(b->get_p0());
                                            return std::move(*this).optimized();
                                        }
                }   }       }   }
        } break;
        case qbcSINH: if(n0) return (new expr_number(std::sinh(n0->get_f()))); break;
        case qbcCOSH: if(n0) return (new expr_number(std::cosh(n0->get_f()))); break;
        case qbcTANH: if(n0) return (new expr_number(std::tanh(n0->get_f()))); break;
        case qbcASINH: if(n0) return (new expr_number(std::asinh(n0->get_f()))); break;
        case qbcACOSH: if(n0) return (new expr_number(std::acosh(n0->get_f()))); break;
        case qbcATANH: if(n0) return (new expr_number(std::atanh(n0->get_f()))); break;

        // INT: Rounds down, no type conversion
        case INT: if(n0) return (new expr_number(std::floor(n0->get_f()))); break;

        // FIX: Rounds towards zero, no type conversion
        case FIX: if(n0) return (new expr_number(std::trunc(n0->get_f()))); break;

        // CINT: Round towards nearest, result is TypeInt
        case CINT: if(n0) return (new expr_number((short)std::round(n0->get_f()), TypeInt)); break;

        // CLNG: Round toward nearest, result is TypeLng
        case CLNG: if(n0) return (new expr_number((int)std::round(n0->get_f()), TypeLng)); break;

        // CSNG & CDBL: Don't do anything in this context
        case CSNG: if(n0) return new expr_number(n0->get_f(), TypeSng); break;
        case CDBL: if(n0) return new expr_number(n0->get_f(), TypeDbl); break;

        /* TODO: Check some formulas used to derive numerical functions:
         *
         *   sinh(x) = 0.5 * (exp(x) - exp(-x))
         *           = 0.5 * exp(x) - 0.5 * exp(-x)
         *
         *   sinh(x)*2 = exp(x) - exp(-x)
         *
         *   cosh(x) = 0.5 * (exp(x) + exp(-x))
         *           = 0.5 * exp(x) + 0.5 * exp(-x)
         *
         *   cosh(x)*2 = exp(x) + exp(-x)
         *
         *   tanh(x*0.5) = (exp(x)-1) / (exp(x)+1)
         *
         *   asinh(x) = log(x + sqrt(x*x + 1))
         *            = log(x + sqrt(x^2 + 1))
         *
         *   acosh(x) = log(x + sqrt(x*x - 1))
         *            = log(x + sqrt(x^2 - 1))
         *
         *   atanh(x)*2 = log( (1+x) / (1-x))
         *              = log( (1+x) / (1-x))
         *
         *   atanh(x) = 0.5 * log(x+1) - 0.5 * log(1-x)
         *
         *   asin(x) = atan(x / sqrt(1 - x*x))
         *           = atan(x / sqrt(1 - x^2))
         *
         *   acos(x) = pi/2 - asin(x)
         */

        case qbcATAN2:
        {
            expr_number* n1 = dynamic_cast<expr_number*> (&*params[1]);
            if(n0&&n1) return new expr_number(std::atan2(n0->get_i(), n1->get_i()));
        } break;

        /*case VAL:
        {
            expr_string* s0 =
                params.size() >= 1
                ? dynamic_cast<expr_string*> (&*params[0])
                : 0;
            if(s0) return (new expr_number(BasVAL(s0->get())));
        }*/
        default: break;
    }
    return std::move(*this).expr::optimized();
}

TypeP expr_builtin_numeric::GetType() const
{
    switch(token)
    {
        case CINT: case CVI: return TypeInt;
        case CLNG: case CVL: return TypeLng;
        case CSNG: case CVS: return TypeSng;
        case CDBL: case CVD: return TypeDbl;
        case SGN: return TypeByte;
        case POS: case CSRLIN: return TypeUint;
        case ASC: case INP: case PEEK: return TypeUbyte;
        case LOC: case LOF: return TypeLng;
        case VAL: return TypeDbl;
        case INSTR: case LEN: return TypeUlong;
        case ABS:
        {
            // Same byte as input, except turned unsigned if possible
            switch(params[0]->GetType()->Type)
            {
                case TypeType::Byte:    case TypeType::UnsignedByte:    return TypeUbyte;
                case TypeType::Integer: case TypeType::UnsignedInteger: return TypeUint;
                case TypeType::Long:    case TypeType::UnsignedLong:    return TypeUlong;
                default: break;
            }
            break;
        }
        default: break;
    }
    // Return TypeSng if none of the inputs to the function are TypeDbl.
    for(const auto& p: params)
        if(p->GetType()->Type == TypeType::Double)
            return TypeDbl;
    return TypeSng;
}

expr* expr_builtin_string::optimized() &&
{
    expr_number* n0 =
        params.size() > 0
        ? dynamic_cast<expr_number*> (&*params[0])
        : 0;

    expr_number* n1 =
        params.size() > 1
        ? dynamic_cast<expr_number*> (&*params[1])
        : 0;

    expr_string* s0 =
        params.size() > 0
        ? dynamic_cast<expr_string*> (&*params[0])
        : 0;

    expr_string* s1 =
        params.size() > 1
        ? dynamic_cast<expr_string*> (&*params[1])
        : 0;

#if 0
        /* Pure functions: functions that only depend on their parameter;
         * can be calculated at compile-time. */
        return (token == INSTR
             || token == MID_S
             || token == MKD_S
             || token == MKI_S
             || token == MKL_S
             || token == MKS_S
               );
#endif
    switch(token)
    {
        case LTRIM_S: if(s0) return (new expr_string(LTrim(s0->get()))); break;
        case RTRIM_S: if(s0) return (new expr_string(RTrim(s0->get()))); break;
        case LCASE_S: if(s0) return (new expr_string(LCase(s0->get()))); break;
        case UCASE_S: if(s0) return (new expr_string(UCase(s0->get()))); break;

        case SPACE_S: if(n0 && n0->is_int())
        {
            if(n0->get_i() <= 64)
                return new expr_string(std::string( n0->get_i(), ' '));
        } break;

        case STRING_S: if(n0 && n0->is_int())
        {
            if(n0->get_i() <= 64)
            {
                if(n1 && n1->get_i() >= 0 && n1->get_i() <= 255)
                    return new expr_string(std::string( n0->get_i(), char(n1->get_i())));
                if(s1 && s1->get().size() >= 1)
                    return new expr_string(std::string( n0->get_i(), s1->get()[0]));
            }
        } break;

        case HEX_S: if(n0 && n0->is_int() && n0->get_i() >= 0)
        {
            std::ostringstream s;  s << std::uppercase << std::hex << n0->get_i();
            return new expr_string(s.str());
        } break;

        case OCT_S: if(n0 && n0->is_int() && n0->get_i() >= 0)
        {
            std::ostringstream s;  s << std::oct << n0->get_i();
            return new expr_string(s.str());
        } break;

        case STR_S: if(n0)
        {
            if(n0->is_int()) return new expr_string(" " + std::to_string(n0->get_i()));
            std::ostringstream s;
            s << ' ' << n0->get_f();
            return new expr_string(s.str());
        } break;

        case CHR_S: if(n0)
        {
            int i = n0->get_i();
            if(i >= 0 && i <= 255) return (new expr_string(std::string(std::size_t(1), char(i))));
        } break;

        case LEFT_S: if(s0 && n1)
        {
            std::size_t i = n1->get_i();
            if(i <= s0->get().size())
                return (new expr_string(s0->get().substr(0, i)));
        } break;

        case RIGHT_S: if(s0 && n1)
        {
            std::size_t i = n1->get_i();
            if(i <= s0->get().size())
                return (new expr_string(s0->get().substr(s0->get().size()-i, i)));
        } break;

        default: break;
    }
    return std::move(*this).expr::optimized();
}

expr_userfun* expr_var::convert_to_functioncall(TypeP type) &&
{
    return new expr_userfun(lexloc, name, type, std::move(indexes));
}


std::string expr::Compile(ExprCompileDefs) const
{
    return "    /* Compilation of " + typenameof(*this) + " not supported */\n";
}

void expr_string::ForEachSubExpr(const expr_visitor& func) const
{
    expr::ForEachSubExpr(func);
}

void expr_number::ForEachSubExpr(const expr_visitor& func) const
{
    expr::ForEachSubExpr(func);
}

void expr_unary::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*p);
    expr::ForEachSubExpr(func);
}

void expr_binary::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*p0);
    func(&*p1);
    expr::ForEachSubExpr(func);
}

void expr_label::ForEachSubExpr(const expr_visitor& func) const
{
    expr::ForEachSubExpr(func);
}

void expr_userfun::ForEachSubExpr(const expr_visitor& func) const
{
    for(std::size_t a=0; a<params.size(); ++a)
        func(&*params[a], maybe, name, a);
    expr::ForEachSubExpr(func);
}

void expr_builtin_function::ForEachSubExpr(const expr_visitor& func) const
{
    for(std::size_t a=0; a<params.size(); ++a)
        func(&*params[a], lvaluemap.size() > a && lvaluemap[a]);
    expr::ForEachSubExpr(func);
}

void expr_caseelse::ForEachSubExpr(const expr_visitor& func) const
{
    expr::ForEachSubExpr(func);
}

void expr_caseis::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*ref);
    expr::ForEachSubExpr(func);
}

void expr_caserange::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*min);
    func(&*max);
    expr::ForEachSubExpr(func);
}

void expr_caseor::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*a);
    func(&*b);
    expr::ForEachSubExpr(func);
}

void expr_var::ForEachSubExpr(const expr_visitor& func) const
{
    for(std::size_t a=0; a<indexes.size(); ++a)
        func(&*indexes[a]);
    // ignore member
    expr::ForEachSubExpr(func);
}

void expr_returnvalueholder::ForEachSubExpr(const expr_visitor& func) const
{
    expr_var::ForEachSubExpr(func);
}

void expr_lubound::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*var);
    func(&*num_index);
    expr::ForEachSubExpr(func);
}

bool expr_string::IsSameAs(const expr& b) const
{
    if(auto e = dynamic_cast<const expr_string*>(&b)) return value == e->value;
    return false;
}
bool expr_number::IsSameAs(const expr& b) const
{
    if(auto e = dynamic_cast<const expr_number*>(&b)) return int_val==e->int_val && float_val==e->float_val;
    return false;
}
bool expr_unary::IsSameAs(const expr& b) const
{
    if(auto e = dynamic_cast<const expr_unary*>(&b)) return token==e->token && p->IsSameAs(*e->p);
    return false;
}
bool expr_binary::IsSameAs(const expr& b) const
{
    if(auto e = dynamic_cast<const expr_binary*>(&b)) return token==e->token && p0->IsSameAs(*e->p0) && p1->IsSameAs(*e->p1);
    return false;
}
bool expr_label::IsSameAs(const expr& b) const
{
    if(auto e = dynamic_cast<const expr_label*>(&b)) return label == e->label;
    return false;
}
bool expr_userfun::IsSameAs(const expr& ) const
{
    return false; // User functions may have side effects
}
bool expr_builtin_function::IsSameAs(const expr& b) const
{
    if(auto e = dynamic_cast<const expr_builtin_function*>(&b))
    {
        switch(token)
        {
            // Only consider functions that we know don't have side effects
            case ABS: case ATN: case CDBL: case CINT: case CLNG: case COS: case SIN: case CSNG:
            case EXP: case FIX: case INT: case LOG: case SQR: case TAN: case ASC:
            case CVD: case CVI: case CVL: case CVS: case VAL: case CSRLIN:
            case ERL: case ERR: case LEN: case VARPTR: case VARSEG:
            // String functions:
            case CHR_S: case HEX_S: case MKD_S: case MKI_S: case MKL_S: case MKS_S:
            case OCT_S: case SPACE_S: case STR_S: case LCASE_S: case UCASE_S:
            case RTRIM_S: case LTRIM_S: case VARPTR_S: case LEFT_S: case RIGHT_S:
            case STRING_S: case MID_S:

                if(params.size() != e->params.size()) return false;
                for(std::size_t a=0; a<params.size(); ++a)
                    if(!params[a]->IsSameAs(*e->params[a])) return false;
                return true;
            default: return false;
        }
    }
    return false;
}
bool expr_var::IsSameAs(const expr& b) const
{
    if(auto e = dynamic_cast<const expr_var*>(&b))
    {
        if(name != e->name) return false;
        if(hasmember() != e->hasmember()) return false;
        if(hasmember() && !getmember().IsSameAs(e->getmember())) return false;
        if(array_as_whole != e->array_as_whole) return false;
        if(indexes.size() != e->indexes.size()) return false;
        for(std::size_t a=0; a<indexes.size(); ++a)
            if(!indexes[a]->IsSameAs(*e->indexes[a])) return false;
        return true;
    }
    return false;
}
bool expr_lubound::IsSameAs(const expr& b) const
{
    if(auto e = dynamic_cast<const expr_lubound*>(&b)) return token==e->token && var->IsSameAs(*e->var) && num_index->IsSameAs(*e->num_index);
    return false;
}

/* Possible expr_binaries are: AND OR XOR  comparisons *+-/ MOD BKSLASH CIRCUMFLEX */
bool expr_binary::MayErr() const
{
    if(p0->MayErr() || p1->MayErr()) return true;
    switch(token)
    {
        // MOD, SLASH and BKSLASH can produce errors if p1can be zero.
        case MOD: case SLASH: case BKSLASH:
            if(auto n = dynamic_cast<const expr_number*> (&*p1))
                return n->get_f() == 0.0;
            return true;

        // CIRCUMFLEX can produce errors if p1 is not integer and p0 is negative.
        case CIRCUMFLEX:
            if(p0->GetType()->is_unsigned()) return false;
            if(p1->GetType()->is_integer()) return false;
            return true;

        default: break;
    }
    return false;
}

bool expr_builtin_function::MayErr() const
{
    for(const auto& p: params)
        if(p->MayErr()) return true;

    switch(token)
    {
        // TODO: qbcASIN and other functions might theoretically error,
        //       but it is currently unimplemented.

        case CHR_S:
            // CHR$() can error if the operand type is anything other than TypeUbyte
            if(params[0]->GetType()->Type == TypeType::UnsignedByte) return false;
            return true;

        case STRING_S:
        case LEFT_S:
        case RIGHT_S:
        case MID_S:
        case INSTR:
        case SPACE_S:
            // These functions may error only if numeric parameters can be negative
            for(const auto& p: params)
                if(p->GetType()->is_num() && !p->GetType()->is_unsigned())
                    return true;
            return false;
        case ASC:
            // ASC can error if the string is empty
            return true;
        case INPUT_S:
        case LPOS:
        case LOC:
        case LOF:
        case FILEATTR:
            // File I/O can always error
            return true;
        case VAL:
            return true;
        case POINT:
        case PMAP:
        case qb64NEWIMAGE:
        case qb64SNDOPEN:
            // Graphics can error
            return true;

        default: break;
    }
    return false;
}

bool expr_lubound::MayErr() const
{
    if(num_index && num_index->MayErr()) return true;

    if(auto v = dynamic_cast<const expr_var*> (&*var))
        if(v->GetType() && (
          v->GetType()->arraytype == TypeInfo::DynamicArray
       || v->GetType()->arraytype == TypeInfo::ParamArray
         )) return true;
    return false;
}

bool expr_var::MayErr() const
{
    /* Note: type->arraytype won't work. It only gives the type of
     *       the individual element from the array.
     *       But, as for the array itself?
     */
    if(basetype)
        switch(basetype->arraytype)
        {
            case TypeInfo::DynamicArray:
            case TypeInfo::ParamArray:
                return true;
            case TypeInfo::StaticArray:
                if(CheckStaticSubscripts)
                    return true;
                break;
            case TypeInfo::NotArray:
                break;
        }
    else
    {
        dprintf(dp_errors, "Internal error: MayErr called on a variable without base type, '%s' (type=%s)\n", name.c_str(),
            type ? type->Dump().c_str() : "(null)");
    }

    for(const auto& i: indexes)
        if(i->MayErr()) return true;

    if(member && member->MayErr()) return true;

    return false;
}
