#include "compile.hh"
#include "parser.hh"

#include <cstdlib>
#include <cstring> // strchr
#include <vector>
#include <string>

#include <unistd.h> // execvp
#include <sys/stat.h> // stat

#include <iostream> // std::cout

#include "dprintf.hh"

static std::string argv0;

static std::string FindQBC_RunningPath()
{
    static char* path = nullptr;
    if(!path)
    {
        path = realpath(argv0.c_str(), NULL);
        char* end = std::strchr(path, '\0');
        while(end > path && *end != '/') { --end; }
        *end = '\0';
    }
    return path;
}

static void AddDirIfExists(std::vector<std::string>& output, const std::string& path)
{
    struct stat st;
    if(stat(path.c_str(), &st)==0 && S_ISDIR(st.st_mode))
        output.push_back(path);
}

static std::vector<std::string> FindIncludePaths()
{
    std::vector<std::string> result;

    AddDirIfExists(result,  FindQBC_RunningPath() + "/qbc" );
    AddDirIfExists(result, "/usr/local/" TARGET "/include/qbc");
    AddDirIfExists(result, "/usr/" TARGET "/include/qbc");
    AddDirIfExists(result, "/usr/include/qbc");
    return result;
}
static std::vector<std::string> FindLibraryPaths()
{
    std::vector<std::string> result;

    AddDirIfExists(result,  FindQBC_RunningPath() + "/qbc");
    AddDirIfExists(result, "/usr/local/" TARGET "/lib/qbc");
    AddDirIfExists(result, "/usr/" TARGET "/lib/qbc");
    AddDirIfExists(result, "/usr/lib/qbc");
    return result;
}

int main(int argc, char** argv)
{
    const char* cxx = std::getenv("CXX");
    if(!cxx) cxx = DEFAULT_CXX;

    const char* cxxflags = std::getenv("CXXFLAGS");
    if(!cxxflags) cxxflags = "";

    const char* cppflags = std::getenv("CPPFLAGS");
    if(!cppflags) cppflags = "";

    argv0 = argv[0];

    std::vector<std::string> args(argv+0, argv+argc);

    std::string opt;
    for(char c: std::string(cppflags))
    {
        if(c==' ') { if(!opt.empty()) { args.push_back(opt); opt.clear(); } continue; }
        opt += c;
    }
    if(!opt.empty()) { args.push_back(opt); opt.clear(); }
    for(char c: std::string(cxxflags))
    {
        if(c==' ') { if(!opt.empty()) { args.push_back(opt); opt.clear(); } continue; }
        opt += c;
    }
    if(!opt.empty()) { args.push_back(opt); opt.clear(); }

    bool expects_option = false;
    std::vector<std::string> gcc_options;
    std::vector<std::string> filenames;
    if(!args.empty()) args.erase(args.begin()); // Remove argv[0]

    bool shared_qblib = false;
    bool verbose_exec = false;
    bool compilation  = true;

    gcc_options.push_back("-std=c++11");
    for(const auto& s: FindIncludePaths()) gcc_options.push_back("-I" + s);
    for(const auto& s: FindLibraryPaths()) gcc_options.push_back("-L" + s);
    gcc_options.push_back("-Wno-missing-field-initializers");
    gcc_options.push_back("-Wno-unused-function");

    for(const auto& s: args)
    {
        if(expects_option && (s[0] != '-' || s=="-"))
        {
            gcc_options.push_back(s);
            expects_option = false;
            continue;
        }
        expects_option = false;

        if(s.empty()) continue;
        if(s=="-h" || s=="--help")
        {
            std::cout <<
               //12345678901234567890123456789012345678901234567890123456789012345678901234567890
                "qbc v" VERSION " - Copyright (C) 2000-2014 Joel Yliluoma (http://iki.fi/bisqwit/)\n"
                "\n"
                "Usage: " << argv[0] << " [<options>] <source file(s)..>\n"
                "\n"
                "Compiles QuickBASIC source code into an executable file through GCC.\n"
                "\n"
                "  --help, -h                  This help\n"
                "  --version, -V               Shows the program version number\n"
                "  -EE                         Prints the generated C++ code without compiling it\n"
                "  -fresume, -fno-resume       Enables/disables support for RESUME\n"
                "                              Default: enabled only if RESUME/ON ERROR are used\n"
                "  -ferl, -fno-erl             Enables/disables support line number information\n"
                "                              Default: enabled only if ERL is used\n"
                "  -fevents, -fno-events       Enables/disables support for event checking\n"
                "                              Default: enabled only if ON KEY/ON TIMER/ON PLAY\n"
                "                                       are used\n"
                "  -fdiagnostics-color=[never|always|auto] Colorize diagnostics\n"
                "  -fcheck-subscripts          Check static array subscripts\n"
                "  -flines, -fno-lines         Enable/disable #line lines in generated C++ code\n"
                "  -fshared-qblib, -fno-static-qblib\n"
                "                              Link qblib shared/dynamically rather than statically\n"
                "  -fstatic-qblib, -fno-shared-qblib\n"
                "                              Link qblib statically rather than dynamically\n"
                "\n"
                " Any other options are passed directly to the compiler.\n"
                " Note that a GCC-compatible compiler is required because of the following\n"
                " two C++ language extensions that are used by QBLIB runtime:\n"
                "   - Label address:         e.g. void* pointer = &&test2;   test2:;\n"
                "   - Computed goto:         e.g. goto *pointer;\n"
                " These extensiona are required to implement RESUME and GOSUB very efficiently.\n"
                "\n"
                " The CXX, CXXFLAGS and CPPFLAGS environment variables are honored.\n"
                " This version of qbc assumes CXX=" << DEFAULT_CXX << " if not specified.\n"
                " libsdl1.2-dev and libslang2-dev must be installed.\n";
            return 0;
        }
        if(s=="-V"||s=="--version")
        {
            std::cout << VERSION"\n";
            return 0;
        }
        if(s[0] != '-' || s=="-")
        {
            filenames.push_back(s);
            continue;
        }
        if(s == "-v") { verbose_exec = true; } // Also pass this option to gcc
        if(s == "-EE") { compilation = false; continue; }
        if(s == "-fresume")    { DoResumeHandling = true; continue; }
        if(s == "-fno-resume") { DoResumeHandling = false; continue; }
        if(s == "-ferl")       { DoERLhandling = true; continue;  }
        if(s == "-fno-erl")    { DoERLhandling = false; continue; }
        if(s == "-fevents")    { DoEventChecks = true; continue;  }
        if(s == "-fno-events") { DoEventChecks = false; continue; }
        if(s == "-flines")     { MakeLineLines = true; continue;  }
        if(s == "-fno-lines")  { MakeLineLines = false; continue; }
        if(s == "-fshared-qblib" || s == "-fno-static-qblib") { shared_qblib = true;  continue; }
        if(s == "-fstatic-qblib" || s == "-fno-shared-qblib") { shared_qblib = false; continue; }
        if(s == "-fcheck-subscripts")    { CheckStaticSubscripts = true; continue;  }
        if(s == "-fno-check-subscripts") { CheckStaticSubscripts = false; continue; }
        if(s == "-fno-diagnostics-color" || s == "-fdiagnostics-color=never") AllowColors=false;
        if(s == "-fdiagnostics-color"    || s == "-fdiagnostics-color=always") AllowColors=true;
        if(s == "-fdiagnostics-color=auto") AllowColors=maybe;

        if(s=="-o" || s=="-I" || s=="-L" || s=="-x"
        || s=="-B" || s=="-F" || s=="-l"
        || s=="-Xlinker" || s=="-Xassembler" ||  s=="-Xpreprocessor"
        || s=="-dumpdir" || s=="-dumpbase" || s=="--param")
        {
            expects_option = true;
        }
        gcc_options.push_back(s);
    }

    if(filenames.size() != 1)
    {
        std::fprintf(stderr, "Error: Expected exactly one filename. Got %u", (unsigned)filenames.size());
        if(!filenames.empty())
        {
            for(std::size_t a=0; a<filenames.size(); ++a)
                std::fprintf(stderr, "%s%s", a ? ", " : " (", filenames[a].c_str());
            std::fprintf(stderr, ")");
        }
        std::fprintf(stderr, "\n");
        return -1;
    }

    std::string bas_filename = filenames[0];

    if(!compilation)
    {
        if(!ParseAndCompile(bas_filename, ""))
            return -1;
        return 0;
    }

    std::size_t slash_pos = bas_filename.rfind('/');
    if(slash_pos == bas_filename.npos)
        slash_pos = 0;
    else
        ++slash_pos;
    std::size_t period_pos = bas_filename.find('.', slash_pos);

    std::string cpp_filename = bas_filename.substr(0, period_pos) + "-bas.cc";

    if(!ParseAndCompile(bas_filename, cpp_filename))
    {
        return -1;
    }

    std::vector<const char*> gcc_args;
    gcc_args.push_back(cxx);
    gcc_args.push_back(cpp_filename.c_str());

    for(const auto& s: gcc_options)
        gcc_args.push_back(s.c_str());

    gcc_args.push_back("-lSDL");
    gcc_args.push_back("-lslang");

    if(!shared_qblib)
    {
        gcc_args.push_back("-Wl,-Bstatic");
        gcc_args.push_back("-lqblib");
        gcc_args.push_back("-Wl,-Bdynamic");
    }
    else
    {
        gcc_args.push_back("-lqblib");
    }

    if(verbose_exec)
    {
        std::fprintf(stderr, "QBC: running");
        for(auto s: gcc_args) std::fprintf(stderr, " %s", s);
        std::fprintf(stderr, "\n");
    }

    gcc_args.push_back(nullptr);

    execvp(cxx, const_cast<char**> (&gcc_args[0]));
    std::perror("execvp");
    return -1;
}
