#include <deque>
#include <set>
#include <cstdio>

template<typename T>
class Lex
{
private:
    static const long eof_pos = -1;

    struct LexResultRec
    {
        LexLocation loc;
        T           tok;
        LexerResult info;
    public:
        LexResultRec(const LexLocation&l, T t)
            : loc(l),tok(t),info() { }
        LexResultRec(const LexLocation&l, T t, const LexerResult& r)
            : loc(l),tok(t),info(r) { }
    };

public:
    class iterator
    {
        Lex* parent;
        long pos;
    public:
        iterator() : parent(0),pos(eof_pos) { }
        iterator(const iterator& ref)
            : parent(ref.parent), pos(ref.pos)
            { if(parent) parent->referrer_list.insert(this); }
        ~iterator()
            { if(parent) {
              parent->referrer_list.erase(this);
              parent->updatepos();
            } }

        const T& operator* () const { return parent->get(pos); }
        iterator& operator++() { inc(); return *this; }
        iterator operator++(int) { iterator tmp(*this); inc(); return tmp; }
        bool operator==(const iterator& b) const { return pos == b.pos; }
        bool operator!=(const iterator& b) const { return !operator==(b); }

        typedef LexLocation LocType;
        typedef LexerResult InfoType;

        const LocType GetLoc() const
        {
            if(!parent) return LexLocation();
            return parent->getloc(pos);
        }
        const InfoType GetInfo() const
        {
            if(!parent) return LexerResult();
            return parent->getinfo(pos);
        }

    private:
        friend class Lex;
        iterator(Lex& obj, long p)
            : parent(&obj), pos(p)
            { parent->referrer_list.insert(this); }

        void inc()
        {
            ++pos;
            parent->updatepos();
            if(parent->eof(pos)) { pos = eof_pos; }
        }
    };

    iterator begin() { return iterator(*this, 0); }
    iterator end()   { return iterator(); }

    Lex()
    {
    }

private:
    bool eof(long pos) const
    {
        return (unsigned)(pos-minimum) >= elements.size() && !supplementable;
    }
    const T& get(long pos) const
    {
#ifdef LEXER_DEBUG
        dprintf(dp_lexer_important, "Asked token %ld: %s\n",
            (long)(pos),
            GetTokenName(  elements[pos-minimum].tok  ).c_str());
#endif
        return elements[pos-minimum].tok;
    }
    const LexLocation& getloc(long pos) const
    {
        return elements[pos-minimum].loc;
    }
    const LexerResult& getinfo(long pos) const
    {
        return elements[pos-minimum].info;
    }
    void updatepos()
    {
        long smallest_pos = 0;
        long biggest_pos = 0;
        bool first = true;
        for(typename std::set<iterator*>::const_iterator
            i = referrer_list.begin();
            i != referrer_list.end();
            ++i)
        {
            if((*i)->pos == eof_pos) continue; // ignore EOFs

            if(first) { first=false; smallest_pos = biggest_pos = (*i)->pos; continue; }
            if((*i)->pos < smallest_pos) smallest_pos = (*i)->pos;
            if((*i)->pos >  biggest_pos)  biggest_pos = (*i)->pos;
        }
#ifdef LEXER_DEBUG
        dprintf(dp_lexer, "Updatepos: smallest=%ld, biggest=%ld (%lu referrers)\n",
            smallest_pos, biggest_pos,
            (unsigned long)referrer_list.size());
#endif
        long excess = smallest_pos - minimum;
        if(excess > 0)
        {
#ifdef LEXER_DEBUG
            dprintf(dp_lexer, "discarding %ld excess tokens\n", excess);
#endif
            elements.erase(elements.begin(), elements.begin() + excess);
            minimum = smallest_pos;
        }

        long need_size = (biggest_pos - minimum) - elements.size() + 1;
        try {
            while(need_size > 0 && supplementable)
            {
                T tok;

                #define Gt() do { \
                     LexResultRec p = GetToken(); \
                     tok=p.tok; \
                     elements.push_back(p); --need_size; \
                   } while(0)

                Gt();

#if 1 /* QuickBasic FOR-NEXT shortcut syntax */
                if(tok == NEXT)
                {
                    Gt();
                    if(tok == IDENTIFIER)
                    {
                        while(supplementable)
                        {
                            Gt();
                            if(tok != COMMA) break;

                            Gt();
                            if(tok != IDENTIFIER) break;

                          //LexResultRec& prev1 = elements[elements.size()-1];
                            LexResultRec& prev2 = elements[elements.size()-2];

                            /* Change the comma to NEXT */
                            prev2.tok = NEXT;

                            /* And insert a COLONCHAR before it */
                            elements.insert(elements.end() - 2,
                                LexResultRec(prev2.loc, COLONCHAR));
                        }
                    }
                }
#endif
            }
        }
        catch(int)
        {
        }
    }

    /* GetToken() : Gets an actual token from qb.lex.l
     *
     * Sets supplementable=true if the file is EOF and further calls to GetToken() are futile.
     *
     * Does not change any iterators or stuff.
     */
    LexResultRec GetToken()
    {
        LexLocation loc = yyGetLocation();

        T tok = LexerGetToken();
#ifdef LEXER_DEBUG
        dprintf(dp_lexer, "Token[%ld]: %s\n",
            (long)(elements.size()+minimum),
            GetTokenName(tok).c_str());
#endif
        if(!tok)
        {
            supplementable = false;
            throw EOF;
        }

        return LexResultRec(loc, tok, yylval);
    }

private:
    /* copies not supported */
    Lex(const Lex&);
    void operator=(const Lex&);

private:
    std::deque< LexResultRec > elements;
    std::set<iterator*> referrer_list;
    long minimum        = 0l;
    bool supplementable = true;
};
