#ifndef bqtQBlibGfxAPIHH
#define bqtQBlibGfxAPIHH

#include <algorithm>
#include <utility>
#include <cstdio> // for size_t
#include <cctype>
#include <math.h> // for sin, cos
#include <cstring> // memset

#include "error.hh"
#include "exec.hh"

class Screen
{
protected:
    int fgc, bgc;
public:
    Screen(): fgc(1), bgc(0) { }

    virtual ~Screen() { }

    /* Terminal specific */
    virtual std::pair<int,int> GetCursorPos() = 0; // x, y
    virtual void SetCursorPos(const std::pair<int,int>& newpos) = 0; // x, y
    virtual void SetCursorVisibility(bool visible) = 0;
    virtual void SetCursorSize(int start, int stop) = 0;

    virtual void SetPages(int npages) = 0;
    virtual void SetAPage(int pageno) = 0;
    virtual void SetVPage(int pageno) = 0;

    virtual std::pair<int,int> GetWinSize() = 0;   // x, y
    virtual std::pair<int,int> ReadScreen(int x,int y) = 0; // chr, color
    virtual void SetWinSize(const std::pair<int,int>& newsize, bool approximate) = 0; // x, y

    virtual void OutS(const char* buffer, size_t nbytes) = 0;
    virtual void ScreenPOKE(int x,int y, int colorflag, int value) = 0;

    virtual void ClearEntireScreen() = 0;
    virtual void ClearText(int firstline, int lastline) = 0;

    virtual void SetBorderColor(int) { }

    virtual void ValidateColor(int color) = 0;

    virtual int GetModeNumber() = 0;

    void SetColor(int fg=-1, int bg=-1, int border=-1)
    {
        if(fg != -1) { ValidateColor(fg); fgc=fg; }
        if(bg != -1) { ValidateColor(bg); bgc=bg; }
        if(border != -1) { ValidateColor(border); SetBorderColor(border); }
    }
    int GetBGC() const { return bgc; }
    int GetFGC() const { return fgc; }

    virtual std::string Inkeys(bool wait) = 0; // INKEY$
    virtual int RawInput() = 0;                // INP(&H60)

    /* View print */
private:
    std::pair<int,int> ViewPrint;
public:
    const std::pair<int,int>& GetViewPrint() const { return ViewPrint; }
    void ResetViewPrint()
    {
        std::pair<int,int> limits = GetWinSize();
        ViewPrint.first  = 1;
        ViewPrint.second = limits.second;
    }
    void SetViewPrint(int from, int to)
    {
        if(from == -1 && to == -1)
        {
            from = 1;
            to = GetWinSize().second;
        }

        ViewPrint.first = from;
        ViewPrint.second = to;
        std::pair<int,int> pos = GetCursorPos();
        /*
        fprintf(stderr, "viewprint(%d,%d); cursor(%d,%d)\n",
            from,to, pos.first,pos.second); fflush(stderr);*/
        if(pos.second < from)      { pos.second = from; SetCursorPos(pos); }
        else if(pos.second >   to) { pos.second = to;   SetCursorPos(pos); }
    }
    void CheckViewPrint(int line)
    {
        if(line < ViewPrint.first || line > ViewPrint.second)
            B_ERROR(5); // illegal function call
    }

    /* CLS */
    void CLS(int type)
    {
/*
    CLS [-1] Clears either the text or graphics viewport. If a graphics
             viewport has been set (using VIEW), clears only the graphics
             viewport. Otherwise, clears the text viewport or entire screen.
    CLS 0    Clears the screen of all text and graphics.
    CLS 1    Clears the graphics viewport or the entire screen if no graphics
             viewport has been set.
    CLS 2    Clears the text viewport.
*/
        if(type == 0)
        {
            ClearEntireScreen();
            return;
        }
        bool has_gfx_viewport = false;
        if((type < 0 && !has_gfx_viewport) || type == 2)
            ClearText(ViewPrint.first, ViewPrint.second);
        else if((type < 0 && has_gfx_viewport) || type == 1)
            ClearGFXport();
    }

    /* Graphics specific */
private:
    std::pair<int,int> GFXcursor;
    std::pair<int,int> GFXresolution;
    std::pair<std::pair<double,double>, std::pair<double,double> > GFXwindow;
    std::pair<std::pair<int,int>, std::pair<int,int> > GFXview;
    std::pair<std::pair<double,double>, std::pair<double,double> > GFXtransformation;
    bool VIEWscreen;
    bool WINDOWscreen;
protected:
    bool GFXwindowSet;
    bool GFXviewSet;
public:
    const std::pair<int,int>& GetGFXcursor() const { return GFXcursor; }
    const std::pair<std::pair<int,int>, std::pair<int,int> >& GetGFXview() const { return GFXview; }
    const std::pair<int,int>& GetGFXresolution() const { return GFXresolution; }

    double GetAspect()
    {
        // this matches what QB does in CIRCLE
        return (4.0 / 3.0) * GFXresolution.second / GFXresolution.first;
    }

    std::pair<int,int> TranslateCoord(bool step,double x,double y)
    {
        std::pair<int,int> result;
        if(step) { x += GFXcursor.first; y += GFXcursor.second; }

        if(GFXwindowSet || GFXviewSet)
        {
            x = x * GFXtransformation.second.first  + GFXtransformation.first.first;
            y = y * GFXtransformation.second.second + GFXtransformation.first.second;
        }
        return std::pair<int,int> ( (int)(x), (int)(y) );
    }
    void UpdateCoordinateTransformation()
    {
        /*
        Method of transforming coordinates:

        VIEWmode WINDOWmode Transformation

        none     none       No transformation
        none     ACTIVE     Window scaled to resolution upsidedown
        none     SCREEN     Window scaled to resolution
        ACTIVE   none       Coordinates offset by VIEW, no scaling
        ACTIVE   ACTIVE     Window scaled to view limits upsidedown
        ACTIVE   SCREEN     Window scaled to view limits
        SCREEN   none       No transformation
        SCREEN   ACTIVE     Window scaled to resolution upsidedown
        SCREEN   SCREEN     Window scaled to resolution
        */
        int result_x1 = 0, result_y1 = 0;
        int result_x2 = GFXresolution.first, result_y2 = GFXresolution.second;

        double source_x1 = 0, source_y1 = 0;
        double source_x2 = result_x2, source_y2 = result_y2;

        if(GFXviewSet && !VIEWscreen)
        {
            if(GFXwindowSet)
            {
                result_x1 = GFXview.first.first;
                result_y1 = GFXview.first.second;
                result_x2 = GFXview.second.first;
                result_y2 = GFXview.second.second;
            }
            else
            {
                result_x1 += GFXview.first.first;
                result_y1 += GFXview.first.second;
                result_x2 += GFXview.second.first;
                result_y2 += GFXview.second.second;
            }
        }

        if(GFXwindowSet)
        {
            source_x1 = GFXwindow.first.first;
            source_y1 = GFXwindow.first.second;
            source_x2 = GFXwindow.second.first;
            source_y2 = GFXwindow.second.second;

            if(WINDOWscreen)
                std::swap(source_y1, source_y2);
        }

        /*
        Now, user-supplied coordinates are supposed
        to be translated as follows:

        xscale = (result_x2 - result_x1) / (source_x2 - source_x1)
        yscale = (result_y2 - result_y1) / (source_y2 - source_y1)
        x1add  = result_x1 - source_x1 * xscale
        y1add  = result_y1 - source_y1 * xscale

        x = x * xscale + x1add;
        y = y * yscale + y1add;
        */
        GFXtransformation.second.first  = (result_x2 - result_x1) / (source_x2 - source_x1);
        GFXtransformation.second.second = (result_y2 - result_y1) / (source_y2 - source_y1);
        GFXtransformation.first.first  = result_x1 - source_x1 * GFXtransformation.second.first;
        GFXtransformation.first.second = result_y1 - source_y1 * GFXtransformation.second.second;
    }
    void SetGFXwindow(bool screen, double x1,double y1, double x2,double y2)
    {
        FailIfNotGraphicsMode();
        if(x1 == x2 || y1 == y2) B_ERROR(5);
        if(x1 > x2) std::swap(x1,x2);
        if(y1 > y2) std::swap(y1,y2);

        WINDOWscreen = screen;
        GFXwindowSet = true;
        GFXwindow.first.first = x1;
        GFXwindow.first.second = y1;
        GFXwindow.second.first = x2;
        GFXwindow.second.second = y2;
        UpdateCoordinateTransformation();
    }
    void SetGFXview(bool screen, int x1,int y1, int x2,int y2)
    {
        FailIfNotGraphicsMode();
        if(x1 == x2 || y1 == y2) B_ERROR(5);
        if(x1 > x2) std::swap(x1,x2);
        if(y1 > y2) std::swap(y1,y2);

        VIEWscreen = screen;
        GFXviewSet = true;
        GFXview.first.first = x1;
        GFXview.first.second = y1;
        GFXview.second.first = x2;
        GFXview.second.second = y2;
        UpdateCoordinateTransformation();
    }
    void ResetGFXwindow()
    {
        FailIfNotGraphicsMode();
        GFXwindowSet = false;
    }
    void ResetGFXview()
    {
        FailIfNotGraphicsMode();
        VIEWscreen = false;
        GFXviewSet = false;
        GFXview.first  = std::pair<int,int>(1,1);
        GFXview.second = GFXresolution;
    }

    void SetNewResolution(int x,int y)
    {
        GFXresolution.first = x;  GFXcursor.first  = x/2;
        GFXresolution.second = y; GFXcursor.second = y/2;
        ResetGFXwindow();
        ResetGFXview();
    }

    struct DrawInfo
    {
        int Recursion;
        int Tangle;
        int Color;
        int Scale;
        bool DontDraw;
        bool MoveAndReturn;
        double msina, cosa, dsina;
    };
    void Draw(const std::string& commands)
    {
        /* Cursor movement:
            Optional prefix:
             B = move, but do not draw
             N = move, and return to original position when done
            Commands:
               HUE
               L R [n] = each of these moves by [n] to the indicated
               GDF       direction

               M x,y   = move absolute or relative
                           +x = move right by x
                           -x = move left by x
                           x = move to absolute x coordinate
                         same for y
          Angle, color and rotation:
             A n = set angle of rotation to n*90 degrees.
                   Note: Figures rotated by 90 or 270 degrees are scaled so
                         that they appear the same size on 4/3 monitor.
             TA n = turn an angle of degrees (-360 to 360)

             C n = set color to n
             S n = set scale to n (1 to 255).
                   Multiplies the values used by move commands.

             P a,b = paint at current spot, filling with a, stopping when b.

          Substring
             X offs = where offs is VARPTR$(std::string object),
                      executes another string.
                      Note: This may cause recursion. QBasic
                      issues an "Out of stack space" error
                      when it happens.
          Value may be given as:
             = offs = where offs is VARPTR$(numeric expression),

                       Note: VARPTR$ strings in QB are three byte strings
                       of the following format:
                             b type
                             w offset
                       Where type is any of: 2% 3$ 4! 8# 20&
                       Obviously we will need a bit more than that.

         */
        DrawInfo info;
        std::memset(&info, 0, sizeof(info));

        info.Color = fgc;
        info.Scale = 1;

        DoDraw(info, commands);
    }
    void DoDraw(DrawInfo& info, const std::string& commands)
    {
        if(++info.Recursion > 50)
        {
            B_ERROR(79); // stack overflow
        }

        int movex=0, movey=0;
        char cmd=0;

        const size_t b=commands.size();
        for(size_t a=0; a<b; )
            switch(cmd = std::toupper(commands[a++]))
            {
                case ' ': case '\t': continue;
                case 'B': info.DontDraw      = true; break;
                case 'N': info.MoveAndReturn = true; break;
                case 'H': movex=-1;movey=-1; goto Move;
                case 'U': movex=0; movey=-1; goto Move;
                case 'E': movex=1; movey=-1; goto Move;
                case 'L': movex=-1;movey=0; goto Move;
                case 'R': movex=-1;movey=0; goto Move;
                case 'G': movex=-1;movey=1; goto Move;
                case 'D': movex=0; movey=1; goto Move;
                case 'F': movex=1; movey=1; goto Move;
                { Move:
                    int sign = 0, amount = DrawGetInt(commands, a, sign) * info.Scale;
                    if(sign != 0) B_ERROR(5);
                    movex *= amount;
                    movey *= amount;

                    double mx = movex, my = movey;
                    if(info.Tangle != 0 && info.Tangle != 360 && info.Tangle != -360)
                    {
                        mx = movex * info.msina + movey * info.cosa;
                        my = movex * info.cosa  + movey * info.dsina;
                    }

                    int endx = (int)(GFXcursor.first  + mx);
                    int endy = (int)(GFXcursor.second + my);

                    if(!info.DontDraw)
                    {
                        Line(GFXcursor.first, GFXcursor.second, endx, endy, info.Color);
                    }
                    if(!info.MoveAndReturn)
                    {
                        GFXcursor.first  = endx;
                        GFXcursor.second = endy;
                    }
                    info.DontDraw = info.MoveAndReturn = false;
                    break;
                }
                case 'M':
                {
                    int signx = 0, signy = 0;
                    movex = DrawGetInt(commands, a, signx); if(signx < 0) movex = -movex;
                    movey = DrawGetInt(commands, a, signy); if(signy < 0) movey = -movey;
                    double mx = movex, my = movey;
                    if(info.Tangle != 0 && info.Tangle != 360 && info.Tangle != -360)
                    {
                        mx = movex * info.msina + movey * info.cosa;
                        my = movex * info.cosa  + movey * info.dsina;
                    }
                    if(signx) mx += GFXcursor.first;  else if(GFXviewSet && !VIEWscreen) mx += GFXview.first.first;
                    if(signy) my += GFXcursor.second; else if(GFXviewSet && !VIEWscreen) my += GFXview.first.second;

                    int endx = (int)(mx);
                    int endy = (int)(my);

                    Line(GFXcursor.first, GFXcursor.second, endx, endy, info.Color);
                    GFXcursor.first  = endx;
                    GFXcursor.second = endy;
                    info.DontDraw = info.MoveAndReturn = false;
                    break;
                }
                case 'T':
                    if(a >= b || std::toupper(commands[a]) != 'A') goto errcase;
                    ++a; // fallthru
                case 'A':
                {
                    int sign = 0, angle = DrawGetInt(commands, a, sign);
                    if(sign > 1) B_ERROR(5);
                    if(sign < 0) angle = -angle;
                    if(cmd == 'A') angle *= 90;
                    if(angle < -360 || angle > 360) B_ERROR(5);
                    info.Tangle = angle;

                    if(info.Tangle != 0 && info.Tangle != 360 && info.Tangle != -360)
                    {
                        const double rad    = angle * 3.141592653 / 360.0;
                        const double aspect = GetAspect();
                        info.cosa  = cos(rad);
                        info.msina = -sin(rad) * aspect;
                        info.dsina =  sin(rad) / aspect;
                    }
                    break;
                }
                case 'C':
                {
                    int sign = 0, color = DrawGetInt(commands, a, sign);
                    if(sign != 0) B_ERROR(5);
                    info.Color = color;
                    break;
                }
                case 'S':
                {
                    int sign = 0, scale = DrawGetInt(commands, a, sign);
                    if(sign != 0 || scale < 1 || scale > 255) B_ERROR(5);
                    info.Scale = scale;
                    break;
                }
                case 'P':
                {
                    int sign1 = 0, color1 = DrawGetInt(commands, a, sign1);
                    int sign2 = 0, color2 = DrawGetInt(commands, a, sign2);
                    if(sign1 || sign2 || color1 < 0 || color2 < 0) B_ERROR(5);
                    Paint(color1, color2);
                    break;
                }
                case 'X':
                {
                    if(a + VARPTR_LENGTH > commands.size()) B_ERROR(5);
                    const std::string& subcommand = VARPTR_ReadStr(commands.c_str()+a);
                    a += VARPTR_LENGTH;
                    DoDraw(info, subcommand);
                    break;
                }
                default:
                errcase:
                    B_ERROR(5);
            }
        --info.Recursion;
    }
    int DrawGetInt(const std::string& commands, size_t& a, int& sign)
    {
        sign=0;
        const size_t b = commands.size();
        while(a < b)
            switch(commands[a])
            {
                case ' ': case '\t': ++a; continue;
                case '-': if(sign) B_ERROR(5); sign=-1; ++a; break;
                case '+': if(sign) B_ERROR(5); sign=+1; ++a; break;
                case '=':
                {
                    ++a;
                    if(a + VARPTR_LENGTH > commands.size()) B_ERROR(5);
                    int n = (int)VARPTR_ReadNum(commands.c_str() + a);
                    a += VARPTR_LENGTH;
                    return n;
                }
                case '0' ... '9':
                {
                    int result = 0;
                    while(a < b && commands[a] >= '0' && commands[a] <= '9')
                        result = result*10 + (commands[a++] - '0');
                    return result;
                }
                default: B_ERROR(5);
            }
        B_ERROR(5);
        return 0;
    }

    static void ClipUtil(int& b1,int& a1, int oa1,int ob1, int oa2,int ob2, int vb1)
    {
        /* FIXME: the 0.5 here is pretty meaningless considering the int division */
        b1=vb1; a1=(int)(oa1+0.5+(vb1-ob1)*(oa2-oa1)/(ob2-ob1));
    }
    static void ClipUtil(int& x1,int& y1,int& x2,int& y2,
                         int vx1,int vy1, int vx2,int vy2,
                         unsigned mode)
    {
        int ox1=x1, oy1=y1, ox2=x2, oy2=y2;
        const unsigned N=1, S=2, W=4, E=8;
        /**/ if(mode&N) {    ClipUtil(y1,x1, ox1,oy1, ox2,oy2, vy1); ContNS:
                       if((mode&(E|W)) && (x1 < vx1 || x1 > vx2))
                         { if(mode & E) goto e; else goto w; } }
        else if(mode&S) {    ClipUtil(y1,x1, ox1,oy1, ox2,oy2, vy2); goto ContNS; }
        else if(mode&W) { w: ClipUtil(x1,y1, oy1,ox1, oy2,ox2, vx1); }
        else if(mode&E) { e: ClipUtil(x1,y1, oy1,ox1, oy2,ox2, vx2); }
    }

    static bool ClipLine(int& x1,int& y1, int& x2,int& y2,
                         int vx1,int vy1, int vx2,int vy2)
    {
        const unsigned N=1, S=2, W=4, E=8;
        unsigned C1=0, C2=0;
        if(y1 < vy1) C1 |= N; else if(y1 > vy2) C1 |= S;
        if(x1 < vx1) C1 |= W; else if(x1 > vx2) C1 |= E;
        if(y2 < vy1) C2 |= N; else if(y2 > vy2) C2 |= S;
        if(x2 < vx1) C2 |= W; else if(x2 > vx2) C2 |= E;
        if(!C1 && !C2) return true; // nothing to clip

        ClipUtil(x1,y1, x2,y2, vx1,vy1, vx2,vy2, C1);
        ClipUtil(x2,y2, x1,y1, vx1,vy1, vx2,vy2, C2);

        return x1 >= vx1 && x1 <= vx2  &&  y1 >= vy1 && y1 <= vy2
            && x2 >= vx1 && x2 <= vx2  &&  y2 >= vy1 && y2 <= vy2;
    }

    virtual void FailIfNotGraphicsMode() { B_ERROR(5); }
    virtual void SetPalette(int index, int value) = 0;
    virtual void SetPixel(int x,int y, unsigned color) = 0;
    virtual void Hline(int x1,int x2, int y, unsigned color) = 0;
    virtual void Vline(int y1,int y2, int x, unsigned color) = 0;
    virtual unsigned GetPixel(int x,int y) = 0;

    void Line(int x1,int y1, int x2,int y2, unsigned color, int style=-1)
    {
        /* Clip line */
        bool still_visible = ClipLine(x1,y1, x2,y2,
            GFXview.first.first,  GFXview.first.second,
            GFXview.second.first, GFXview.second.second);
        if(still_visible)
            ClippedLine(x1,y1, x2,y2, color,style);
    }
    void ClippedLine(int x1,int y1, int x2,int y2, unsigned color, int style)
    {
        /* Now we don't have to worry about clipping */

        /* Fill the whole int with the style (originally 16-bit) */
        style *= (~0u / 65535u); // 0x000100010001...

        int deltax = abs(x2-x1), deltay = abs(y2-y1);
        bool steep = deltay > deltax;

        if(steep)   { std::swap(x1,y1); std::swap(x2,y2); std::swap(deltax,deltay); }
        if(x1 > x2) { std::swap(x1,x2); std::swap(y1,y2); }

        int error = (deltax + 1) / 2;
        int ystep = y1 < y2 ? 1 : -1;

        if(style == -1)
        {
            /* Optimize as a set of vertical or horizontal lines */
            int x = x1, npixels = (x2-x1)+1;

            void (Screen::*Stripe)(int x1,int x2,int y,unsigned color)
                = steep ? &Screen::Hline : &Screen::Vline;

            // FIXME: Division by zero if deltay==0

            for(int y=y1;;)
            {
                int seq_length = error / deltay;
                if(seq_length >= npixels)
                {
                    (this->*Stripe)(x,x+npixels-1, y, color);
                    break;
                }
                (this->*Stripe)(x,x+seq_length-1, y, color);
                error   -= seq_length * deltay;
                x       += seq_length;
                npixels -= seq_length;
                y       += ystep;
                error   += deltax;
            }
        }
        else
        {
            const unsigned shiftbits = sizeof(style)*8-1;
            const unsigned topmask   = 1U << shiftbits;
            #define LineSetP(x,y) do { \
                if(style & topmask) SetPixel(x, y, color); \
                style = (style << 1) | ((style&topmask) >> shiftbits); \
            } while(0)
            if(steep)
            {
                // y and x are swapped
                for(int x = x1, y = y1; x <= x2; ++x)
                {
                    LineSetP(y,x);
                    error -= deltay;
                    if(error <= 0) { y += ystep; error += deltax; }
                }
            }
            else
            {
                for(int x = x1, y = y1; x <= x2; ++x)
                {
                    LineSetP(x,y);
                    error -= deltay;
                    if(error <= 0) { y += ystep; error += deltax; }
                }
            }
            #undef LineSetP
        }
    }

    void FloodFill_Recurse(int x,int y, unsigned color,int edge, unsigned forbid_dirs=0)
    {
        const unsigned N=1, S=2, W=4, E=8;

        SetPixel(x,y, color);

        int up = y-1, down=y+1,
          left = x-1, right=x+1;

        /* FIXME: Verify whether this algorithm works as it should */
        if(left >= GFXview.first.first && !(forbid_dirs & W))
        {
            int lp = GetPixel(left,y);
            if(lp != edge)
                FloodFill_Recurse(left,y, color,edge, E);
        }
        if(right <= GFXview.second.first && !(forbid_dirs & E))
        {
            int lp = GetPixel(right,x);
            if(lp != edge)
                FloodFill_Recurse(right,y, color,edge, W);
        }
        if(up >= GFXview.first.second && !(forbid_dirs & S))
        {
            int lp = GetPixel(x,up);
            if(lp != edge && lp != color)
                FloodFill_Recurse(x,up, color,edge, S);
        }
        if(down <= GFXview.second.second && !(forbid_dirs & N))
        {
            int lp = GetPixel(x,down);
            if(lp != edge && lp != color)
                FloodFill_Recurse(x,down, color,edge, N);
        }
    }

    void Paint(unsigned color, int edge)
    {
        // Paints by the given color starting at GFXcursor.
        ValidateColor(color);
        ValidateColor(edge);

        int x = GFXcursor.first, y = GFXcursor.second;
        if(x < GFXview.first.first  || x > GFXview.second.first
        || y < GFXview.first.second || y > GFXview.second.second) return;

        if(GetPixel(x,y) == edge) return;
        FloodFill_Recurse(x,y, color,edge);
    }

    void Paint(const std::string& tile, int edge, int back1=-1, int back2=-1)
    {
        // Paints by the given pattern starting at GFXcursor.
        // For 1bpp modes, tile is 8 pixels wide. (Max: 8 pixels tall)
        // For 2bpp modes, tile is 4 pixels wide. (Max: 16 pixels tall)
        // For 4bpp modes, tile is 2 pixels wide. (Max: 32 pixels tall)
        // For 8bpp modes, tile is 1 pixels wide. (Max: 64 pixels tall)
        // It may consist of up to 64 bytes
        // FIXME: This does not match observation: in 4bpp, tile was repeated horizontally

        int x = GFXcursor.first, y = GFXcursor.second;
        if(x < GFXview.first.first  || x > GFXview.second.first
        || y < GFXview.first.second || y > GFXview.second.second) return;

        ValidateColor(edge);
        // TODO
    }

    virtual void FilledRectangle(int x1,int y1, int x2,int y2, unsigned color)
    {
        ValidateColor(color);
        if(x1 < x2) std::swap(x1, x2);
        if(y1 < y2) std::swap(y1, y2);
        if(x1 < GFXview.first.first) x1 = GFXview.first.first;
        if(x2 > GFXview.second.first) x2 = GFXview.second.first;
        if(y1 < GFXview.first.second) y1 = GFXview.first.second;
        if(y2 > GFXview.second.second) y2 = GFXview.second.second;
        for(int y=y1; y<=y2; ++y)
            Hline(x1,x2, y, color);
    }

    void Rectangle(int x1,int y1, int x2,int y2, unsigned color, int style)
    {
        Line(x1,y1, x2,y1, color,style);
        Line(x1,y2, x2,y2, color,style);
        Line(x1,y1, x1,y2, color,style);
        Line(x2,y1, x2,y2, color,style);
    }

    class CircleRenderer
    {
        bool scaling_x;
        int integ_aspect;    // aspectr
        unsigned num_points; // cnpnts

        int centerline_from_start; //0=no need, 1=will need, 2=done
        int centerline_from_end;   // -"-
        int centerx, centery;

        bool angles_specified;
        bool plot_when_between; // cplotf
        int startcount; // cstcnt
        int endcount;   // cencnt

        int radius;

        Screen* scrn;
    public:
        CircleRenderer(int cx,int cy, int rad, Screen* s)
            : scaling_x(false),
              integ_aspect(256.0),
              num_points( (int)(0.5 + radius * (sqrt(2) / 2.0)) ),
              centerline_from_start(0),
              centerline_from_end(0),
              centerx(cx), centery(cy),
              angles_specified(false),
              plot_when_between(false), startcount(0), endcount(0),
              radius(rad),
              scrn(s)
        {
            SetAspect( s->GetAspect() );
        }

        void SetBeginAngle(double start)
        {
            angles_specified = true;
            endcount = 65535; // infinite end count
            startcount = AngleCalc(start, centerline_from_start);
            plot_when_between = true; // polarity counterclockwise
        }
        void SetEndAngle(double end)
        {
            angles_specified = true;
            endcount = AngleCalc(end, centerline_from_end);
            plot_when_between = true; // polarity counterclockwise
        }
        void SetAspect(double aspect)
        {
            scaling_x = aspect > 1.0;
            if(scaling_x) aspect = 1.0 / aspect;

            integ_aspect = int(256.0 / aspect);
            if((integ_aspect & 0xFF) == 0 && (integ_aspect & 0xFF00))
                integ_aspect = 0x100 + (integ_aspect & 0xFF);
            else
                integ_aspect &= 0xFF; // aspectr
        }
        void Render(unsigned color)
        {
            if(radius < 0) B_ERROR(5);
            scrn->ValidateColor(color);

            // if window scaling is used, radius will be scaled by X coordinate
            //   i.e. multiplied by viewwidth / windowwidth

            // If aspect = 1:
            //   width=radius
            //  height=radius

            // If aspect <= 0:
            //   width=radius
            //   height=radius * (1-frac(abs(aspect)))

            // If aspect < 1
            //   width=radius
            //  height=radius * aspect

            // If aspect > 1
            //   width=radius / aspect
            //  height=radius

            if(startcount >= endcount)
            {
                std::swap(startcount, endcount);
                std::swap(centerline_from_start, centerline_from_end);
                plot_when_between = false; // polarity clockwise?
            }
            if(startcount == endcount
            && (centerline_from_start || centerline_from_end))
                centerline_from_start = centerline_from_end = 1; // set both checks

            int x = radius, y = 0, sum = 1 - radius;
            for(;; ++y)
            {
                CirclePlot8(x,y, color);
                if(y >= x) break;
                if(sum > 0) { sum -= 2*(x-1); }
                else        { sum += 2*y + 3; }
            }
        }
    private:
        bool CenterLine(int pointx,int pointy, int& flag, unsigned color) const
        {
            if(flag == 0) return true; // normal point
            if(flag == 2) return false; // already plotted
            flag = 2;
            scrn->Line(centerx, centery, pointx,pointy, color);
            return false; // don't need to plot anymore
        }

        void CirclePlot4(unsigned octantpointcount,
                         unsigned circlepointcount,
                         int x_1, int y_1, int x_2, int y_2,
                         unsigned color)
        {
            for(unsigned a=4; ; )
            {
                int pointx = x_1 + centerx;
                int pointy = y_1 + centery;

                bool do_plot = true;

                if(angles_specified)
                {
                    int bx = circlepointcount;
                    circlepointcount = num_points * 2 + bx;
                    bx += octantpointcount;

                    if(bx == startcount) // at starting point?
                    {
                        // line at starting point
                        do_plot = CenterLine(pointx,pointy, centerline_from_start, color);
                    }
                    else if(bx == endcount) // at end point?
                    {
                        // line at ending point
                        do_plot = CenterLine(pointx,pointy, centerline_from_end, color);
                    }
                    else
                        do_plot = (bx >= startcount && bx < endcount)
                                ? plot_when_between
                                : !plot_when_between;
                }
                if(do_plot)
                {
                    scrn->SetPixel(pointx,pointy, color);
                }
                if(--a == 0) break;

                std::swap(x_1,y_2); x_1=-x_1;
                std::swap(y_1,x_2); y_1=-y_1;
            }
        }
        void CirclePlot8(int x,int y, unsigned color)
        {
            unsigned octantpointcount = y; // cpcnt
            unsigned circlepointcount = 0; // cpcnt8

            int y_scaled = (y * integ_aspect + 128) >> 8;
            int x_scaled = (x * integ_aspect + 128) >> 8;

            if(scaling_x)
                { std::swap(x, x_scaled);
                  std::swap(y, y_scaled); }

            int x_1 = x_scaled, y_1 = y; // dx>si,    di>di
            int y_2 = y_scaled, x_2 = x; // bx>cxoff, si>cyoff

            CirclePlot4(octantpointcount,
                        circlepointcount,
                        x_1, -y_1,  x_2, y_2, color); // (x,-sy), (-y,-sx), (-x,sy), (y,-sx)
             /*
                      0    1    2    3
              x_1:   x_1 -y_2 -x_1  y_2
              y_1:  -y_1 -x_2  y_1  x_2
              x_2:   x_2 -y_1 -x_2  y_1
              y_2:   y_2  x_1 -y_2 -x_1
             */

            circlepointcount = num_points;
            octantpointcount = circlepointcount - octantpointcount;

            CirclePlot4(octantpointcount,
                        circlepointcount,
                        y_2, -x_2, y_1,  x_1, color); // (y,-sx), (-x,-sx), (-y,sx), (x,sy)
             /*
                      0    1    2    3
              x_1:   y_2 -x_1  -y_2  x_1
              y_1:  -x_2 -y_1   x_2  y_1
              x_2:   y_1 -x_2  -y_1  x_2
              y_2:   x_1  y_2  -x_1 -y_2
             */
        }

        int AngleCalc(double angle, int& centerline)
        {
            const double pi    = M_PI;
            const double pi2   = 2*pi;
            const double pirp4 = 4/pi;

            if(angle < 0) { centerline=1; angle=-angle; }
            if(angle > pi2) B_ERROR(5);

            int octant = 1 + (int)(angle * pirp4);

            int scaledcountbase = (octant &~ 1) * num_points;

            double angle_val = fabs((octant & 2) ? cos(angle) : cos(angle));
            return (int)(0.5 + radius*angle_val + scaledcountbase);
        }
    };

    virtual void ClearGFXport()
    {
        FilledRectangle( GFXview.first.first, GFXview.first.second,
                         GFXview.second.first, GFXview.second.second,
                         bgc );
    }
};

#endif
