#include <ostream>
#include <sstream>

#include "typenameof.hh"
#include "stmt.hh"

stmt::stmt(const LexLocation& ll): lexloc(ll)
{
}
stmt::~stmt()
{
}

void stmt::set_prev(stmtp&& p)
{
    prev = std::move(p);

    // As a special optimization, check if both this and the previous statement
    // is a string-assign statement, and they both address the same variable.
    // This converts the following paradigm:
    //    s = a
    //    s = s + b
    // Into s = (a + b)
    if(prev)
    {
        auto let0 = dynamic_cast<stmt_let*> (&*prev);
        auto let1 = dynamic_cast<stmt_let*> (this);
        if(let0 && let1 && let0->get_target()->IsSameAs(*let1->get_target()))
        {
            const expr* let = &*let1->get_target();
            auto eb = dynamic_cast<expr_binary*> (&*let1->get_value());
            if(eb && eb->p0->IsSameAs(*let))
            {
                // However, we cannot do this if eb->p1 contains references to eb->p0.
                bool contains_ref = false;
                std::function<void(const class expr*,tristate,const std::string&,unsigned)> AnalyzeExpr
                    = [&](const class expr* expr, tristate, const std::string& , unsigned)
                    {
                        if(!expr || contains_ref) return;
                        if(let->IsSameAs(*expr)) { contains_ref=true; return; }
                        expr->ForEachSubExpr(AnalyzeExpr);
                    };
                AnalyzeExpr(&*eb->p1,false,{},0);
                if(!contains_ref)
                {
                    eb->p0 = std::move(let0->get_value());
                    OptimizeExprp(let1->get_value());
                    return set_prev( std::move(let0->prev) );
                }
            }
        }
    }
}

std::string stmt::Compile() const
{
    return "    /* Compilation of " + typenameof(*this) + " not supported */\n";
}

static bool IsEntirelyIntegerCase(expr* exp)
{
    if(!exp) return false;

    if(dynamic_cast<expr_caseelse*>(exp))
    {
        return true;
    }
    if(expr_caseor* e = dynamic_cast<expr_caseor*>(exp))
    {
        return IsEntirelyIntegerCase(e->a.get()) && IsEntirelyIntegerCase(e->b.get());
    }
    if(expr_caseis* e = dynamic_cast<expr_caseis*>(exp))
    {
        if(e->token != EQUALS) return false;
        return IsEntirelyIntegerCase(e->ref.get());
    }
    if(expr_caserange* e = dynamic_cast<expr_caserange*>(exp))
    {
        return IsEntirelyIntegerCase(e->min.get()) && IsEntirelyIntegerCase(e->max.get());
    }

    if(!exp->is_const()) return false;

    if(expr_number* e = dynamic_cast<expr_number*>(exp))
        if(e->is_int()) return true;

    TypeP type = exp->GetType();
    return type && type->is_integer();
}
static bool IsEntirelyStringCase(expr* exp)
{
    if(!exp) return false;

    if(dynamic_cast<expr_caseelse*>(exp))
    {
        return true;
    }
    if(auto e = dynamic_cast<expr_caseor*>(exp))
    {
        return IsEntirelyStringCase(e->a.get()) && IsEntirelyStringCase(e->b.get());
    }
    if(auto e = dynamic_cast<expr_caseis*>(exp))
    {
        if(e->token != EQUALS) return false;
        return IsEntirelyStringCase(e->ref.get());
    }
    if(/***/ dynamic_cast<expr_caserange*>(exp))
    {
        return false;
    }

    if(/***/ dynamic_cast<expr_string*>(exp))
        return true;

    return false;
}

bool stmt_select::IsEntirelyGotoCase() const
{
    for(const auto& c: cases)
    {
        const stmtp& s = c.second;
        const stmt_labelref* l = dynamic_cast<const stmt_labelref*> (&*s);
        if(!l || l->token != GOTO) return false;
    }
    return true;
}

bool stmt_select::IsEntirelyGosubCase() const
{
    for(const auto& c: cases)
    {
        const stmtp& s = c.second;
        const stmt_labelref* l = dynamic_cast<const stmt_labelref*> (&*s);
        if(!l || l->token != GOSUB) return false;
    }
    return true;
}

void stmt_print::OptimizeTail()
{
    std::string buffered_string;
    while(!params.empty())
    {
        std::pair<QBtoken,exprp>& last = params.back();
        if(last.first != (QBtoken)0) break;

        if(expr_string* str = dynamic_cast<expr_string*>(&*last.second))
        {
            buffered_string.insert(0, str->get());
            params.erase(--params.end());
        }
        else if(expr_number* num = dynamic_cast<expr_number*>(&*last.second))
        {
            if(!num->is_int()) break;
            buffered_string.insert(0, std::to_string(num->get_i()));
            params.erase(--params.end());
        }
        else break;
    }
    if(!buffered_string.empty())
        params.emplace_back( (QBtoken)0, std::unique_ptr<expr_string>(new expr_string(buffered_string)) );
}

bool stmt_select::IsEntirelyIntegerCase() const
{
    for(const auto& c: cases)
    {
        if(!::IsEntirelyIntegerCase(c.first.get())) return false;
    }
    return true;
}

bool stmt_select::IsEntirelyStringCase() const
{
    for(const auto& c: cases)
    {
        if(!::IsEntirelyStringCase(c.first.get())) return false;
    }
    return true;
}

void stmt_if::AddBody(stmtp&& thp, std::list<std::pair<exprp, stmtp>>&& elip, stmtp&& elsp)
{
    if(elip.empty() && !thp && elsp)
    {
        // Convert if(x) {} else {stmt}
        //    into if(!x) {stmt}
        std::swap(thp, elsp);
        condition.reset( new expr_unary(NOT, std::move(condition)) );
        OptimizeExprp(condition);
    }

    thenp = std::move(thp);
    if(elip.empty())
    {
        elsep = std::move(elsp);
    }
    else
    {
        // convert if(x) AA elseif(y) BB elseif(z) CC else DD
        //    into if(x) AA else { if(y) BB elseif(z) CC else DD }
        exprp slave_condition = std::move(elip.begin()->first);
        stmtp slave_thenp     = std::move(elip.begin()->second);
        elip.pop_front();

        // FIXME: Proper lex-location
        stmtp slave_if ( new stmt_if(GetLoc(), std::move(slave_condition)) );
        ((stmt_if&)*slave_if).AddBody( std::move(slave_thenp), std::move(elip), std::move(elsp) );

        elsep = std::move(slave_if);
    }
}

void stmt_if::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    func(&*thenp);
    func(&*elsep);
    stmt::ForEachSubStmt(func);
}
void stmt_if::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*condition);
    stmt::ForEachSubExpr(func);
}

void stmt_select::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    for(const auto& c: cases)
        func(&*c.second);
    stmt::ForEachSubStmt(func);
}
void stmt_select::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*ref);
    for(const auto& c: cases)
        func(&*c.first);
    stmt::ForEachSubExpr(func);
}

void stmt_doloop::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    func(&*bodyp);
    stmt::ForEachSubStmt(func);
}
void stmt_doloop::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*precondition);
    func(&*postcondition);
    stmt::ForEachSubExpr(func);
}

void stmt_for::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    func(&*bodyp);
    stmt::ForEachSubStmt(func);
}
void stmt_for::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*var, true);
    func(&*begin);
    func(&*end);
    func(&*step);
    stmt::ForEachSubExpr(func);
}


void stmt_labelref::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_labelref::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*label);
    stmt::ForEachSubExpr(func);
}

void stmt_let::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_let::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*target, true);
    func(&*value);
    stmt::ForEachSubExpr(func);
}

void stmt_print::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_print::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*target);
    func(&*using_string);
    for(const auto& p: params)
        func(&*p.second);
    stmt::ForEachSubExpr(func);
}

void stmt_input::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_input::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*source);
    func(&*prompt);
    for(const auto& p: params)
        func(&*p, true);
    stmt::ForEachSubExpr(func);
}


void stmt_fileopen::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_fileopen::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*filename);
    func(&*mode_str);
    func(&*reclen);
    stmt::ForEachSubExpr(func);
}

void stmt_redim::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_redim::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*target, true);
    for(const auto& d: dimensions)
    {
        func(&*d.first);
        func(&*d.second);
    }
    stmt::ForEachSubExpr(func);
}


void stmt_erase::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_erase::ForEachSubExpr(const expr_visitor& func) const
{
    func(&*target, true);
    stmt::ForEachSubExpr(func);
}

void stmt_subcall::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_subcall::ForEachSubExpr(const expr_visitor& func) const
{
    for(std::size_t a=0; a<params.size(); ++a)
    {
        func(&*params[a], maybe, name, a);
    }
    stmt::ForEachSubExpr(func);
}


void stmt_label::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_label::ForEachSubExpr(const expr_visitor& func) const
{
    stmt::ForEachSubExpr(func);
}


void stmt_lib_stmt::ForEachSubStmt(const std::function<void(const stmt*)>& func) const
{
    stmt::ForEachSubStmt(func);
}
void stmt_lib_stmt::ForEachSubExpr(const expr_visitor& func) const
{
    // lvaluemap defines which values may be assigned to.
    for(std::size_t a=0; a<params.size(); ++a)
    {
        func(&*params[a], a < lvaluemap.size() && lvaluemap[a]);
    }
    stmt::ForEachSubExpr(func);
}
